import java.util.Vector; 

/* Copyright K. Lano 2014
   Module: ATL */ 

public class ATLModule
{ String name; 
  Vector elements; // MatchedRule
  java.util.Map interp = null; 
  Vector attributes = new Vector(); // of Attribute

  public ATLModule(String nme)
  { name = nme; } 

  public void addElement(ModuleElement me) 
  { elements.add(me); }

  public void setElements(Vector elems) 
  { elements = elems; }

  public void addElements(Vector elems) 
  { elements.addAll(elems); }

  public void addAttribute(Attribute att)
  { attributes.add(att); } 

  public void setInterpretation(java.util.Map intp)
  { interp = intp; } 

  public String toString()
  { String res = "module " + name + ";\n"; 
    res = res + "create OUT : T from IN : S;\n";

    for (int i = 0; i < attributes.size(); i++) 
    { Attribute att = (Attribute) attributes.get(i); 
      String attname = att.getName(); 
      Entity ent = att.getEntity(); 
      Type t = att.getType();
      Expression e = att.getInitialExpression();  
      res = res + "  helper "; 
      if (ent != null) 
      { res = res + "context " + ent; } 
      res = res + " def : " + attname + " : " + t + " = " + e + ";\n"; 
    } 

    for (int i = 0; i < elements.size(); i++) 
    { ModuleElement me = (ModuleElement) elements.get(i); 
      res = res + me + "\n"; 
    } 

    return res; 
  }  
     
  public UseCase toUML(Vector types, Vector entities, Vector inits)
  { UseCase res = new UseCase(name,null);
    res.addPostconditions(inits); 

    java.util.Map interp = getInterpretation(); 
    for (int i = 0; i < elements.size(); i++) 
    { ModuleElement me = (ModuleElement) elements.get(i);
      if (me instanceof MatchedRule) 
      { MatchedRule rule = (MatchedRule) me; 
        rule.typeCheck(types,entities); 
        if (rule.isLazy())
        { rule.toOperation(types,entities,interp); } 
        else 
        { Constraint con = rule.toConstraint(types,entities,interp); 
          res.addPostcondition(con);
        }  
      } 
    } 
 
    for (int j = 0; j < attributes.size(); j++) 
    { Attribute att = (Attribute) attributes.get(j); 
      Entity ent = att.getEntity(); 
      if (ent != null) 
      { Expression ini = att.getInitialExpression(); 
        Expression resexp = new BasicExpression("result"); 
        resexp.setType(att.getType()); 
        resexp.setUmlKind(Expression.VARIABLE); 
        Expression post = new BinaryExpression("=", resexp, ini); 
        BehaviouralFeature bf = 
          new BehaviouralFeature(att.getName(), new Vector(), true, att.getType()); 
        ent.addOperation(bf); 
        bf.setPost(post); 
        bf.setEntity(ent); 
      }  
      else 
      { res.addAttribute(att); }  
    } 
    // stat.typeCheck(types,entities,contexts,newparms);
    return res; 
  } 

  public java.util.Map getInterpretation()
  { java.util.Map res = new java.util.HashMap(); 
    for (int i = 0; i < elements.size(); i++) 
    { if (elements.get(i) instanceof MatchedRule) 
      { MatchedRule mr = (MatchedRule) elements.get(i); 
        OutPattern opatt = mr.outPattern; 
        InPattern inpatt = mr.inPattern; 
        if (inpatt != null && opatt != null) 
        { String t1 = inpatt.firstType(); 
          if (t1 != null) 
          { Vector opes = opatt.elements; 
            for (int j = 0; j < opes.size(); j++) 
            { OutPatternElement ope = (OutPatternElement) opes.get(j); 
              Entity tt = ope.getEntity(); 
              res.put(t1,tt);   // No: the common superclass of tt and res(t1) if res(t1) != null
            } 
          }
        } 
      } 
    }
    interp = res; 
    return res; 
  }  

  public Vector dataAnalysis()
  { // returns list of rules that must be split
    Vector res = new Vector(); 
    for (int i = 0; i < elements.size(); i++) 
    { if (elements.get(i) instanceof MatchedRule) 
      { MatchedRule mr = (MatchedRule) elements.get(i);
        if (mr.isLazy()) { continue; } 
        // mr must be split if it reads a target type that is an output of itself or a 
        // later rule.  
        OutPattern opatt = mr.outPattern; 
        Vector streads = opatt.sourceTypesRead(); 

        Statement stat = mr.actionBlock; 
        if (stat != null) 
        { streads.addAll(stat.readFrame()); } 

        System.out.println("Read frame of rule " + i + " is: " + streads); 

        for (int j = i; j < elements.size(); j++) 
        { MatchedRule mr2 = (MatchedRule) elements.get(j); 
          if (!mr2.isLazy() && mr2.hasOutputType(streads))
          { System.err.println("Rule " + mr2 + " writes an entity read by earlier rule " + mr); 
            System.err.println("It will be split into two rules/constraints"); 
            res.add(mr.slice()); 
          } 
        } 
      } 
    } 
    return res; 
  } 
}