import java.util.Vector; 

public class EtlModule
{ String name;
  Vector pre = new Vector();  // of NamedBlock
  Vector post = new Vector(); 
  Vector transformationRules = new Vector();  // of TransformationRule

  public EtlModule(String nme)
  { name = nme; }

  public void addPre(NamedBlock p)
  { pre.add(p); }

  public void addPost(NamedBlock p)
  { post.add(p); }

  public void addRule(TransformationRule r)
  { // if (r.isPrimary())
    // { transformationRules.add(0,r); }
    // else 
    { transformationRules.add(r); }
  }

  public void add(Object ob)
  { if (ob instanceof TransformationRule)
    { addRule((TransformationRule) ob); } 
    else if (ob instanceof NamedBlock)
    { NamedBlock nb = (NamedBlock) ob; 
      if (nb.isPre())
      { addPre(nb); } 
      else 
      { addPost(nb); } 
    } 
  } 


  public String toString()
  { String res = "module " + name + "\n"; 
    for (int i = 0; i < pre.size(); i++)
    { res = res + pre.get(i); }
    for (int i = 0; i < transformationRules.size(); i++)
    { res = res + transformationRules.get(i); }
    for (int i = 0; i < post.size(); i++)
    { res = res + post.get(i); }
    return res + "\n";
  }

  public Vector rulesOnInput(Entity e)
  { Vector res = new Vector();
    for (int i = 0; i < transformationRules.size(); i++)
    { TransformationRule r = (TransformationRule) transformationRules.get(i);
      Attribute att = r.getSource();
      Type atype = att.getType(); 
      if (atype != null && atype.isEntity())
      { Entity src = atype.getEntity(); 
        if (e == src || Entity.isAncestor(e,src) ||
            Entity.isAncestor(src,e)) 
        { res.add(r); }
      } 
    }
    return res;
  }

  public Vector rulesMapping(Entity src, Entity trg)
  { // Rules that can map input x : src to a y : trg
    Vector pres = new Vector();
    Vector res = new Vector();
    Vector srcrules = rulesOnInput(src);
    for (int i = 0; i < srcrules.size(); i++)
    { TransformationRule r = (TransformationRule) srcrules.get(i);
      if (r.hasOutputType(trg))
      { if (r.isPrimary())
        { pres.add(r); }
        else 
        { res.add(r); }
      }
    } // add primary ones first
    pres.addAll(res);
    return pres;
  }

  public UseCase toUseCase(Vector entities, Vector types)
  { UseCase uc = new UseCase(name,null); 
    for (int i = 0; i < pre.size(); i++) 
    { NamedBlock nb = (NamedBlock) pre.get(i); 
      Constraint cons = nb.toConstraint(entities, types); 
      uc.addPostcondition(cons); 
    } 
    for (int i = 0; i < transformationRules.size(); i++)
    { TransformationRule tr = (TransformationRule) transformationRules.get(i);
      if (tr.isLazy()) { } 
      else 
      { Constraint con = tr.toConstraint(entities, types);
        uc.addPostcondition(con); 
      }
    } 
    for (int i = 0; i < post.size(); i++)
    { NamedBlock pb = (NamedBlock) post.get(i);
      Constraint cons = pb.toConstraint(entities, types); 
      uc.addPostcondition(cons); 
    }
    return uc; 
  } 

}

