import java.util.Vector; 

/* Copyright K. Lano, 2014
   Package: ATL */ 

public class MatchedRule extends Rule
{ boolean isLazy = false; 
  boolean isUnique = false; 
  InPattern inPattern = null; 

  public MatchedRule(boolean isL, boolean isU)
  { isLazy = isL; 
    isUnique = isU; 
  } 

  public boolean isLazy() 
  { return isLazy; } 

  public void setInPattern(InPattern inp)
  { inPattern = inp; } 

  public String toString()
  { String res = "rule " + name + " {\n" + 
    "  " + inPattern + "\n" + 
    "  " + outPattern + "\n"; 
    if (isLazy) 
    { res = "lazy " + res; } 
    if (actionBlock != null) 
    { res = res + "  do { " + actionBlock + " }\n"; }  
    res = res + "}"; 
    return res; 
  } 

  public boolean typeCheck(Vector types, Vector entities)
  { Vector env = inPattern.allVariables(); 
    Vector contexts = new Vector(); 
    env.addAll(outPattern.allVariables()); 

    // ante.typeCheck(types,ents,contexts,env);
    if (actionBlock != null) 
    { actionBlock.typeCheck(types,entities,contexts,env); } 
    return true; 
  }  


  public void toOperation(Vector types, Vector ents, java.util.Map interp)
  { // In the case of lazy rules

    Expression ante = inPattern.toExpression(); 
    if (ante == null) 
    { ante = new BasicExpression("true"); } 
    
    Vector env = inPattern.allVariables(); 
    Vector contexts = new Vector(); 
    ante.typeCheck(types,ents,contexts,env);

    Entity owner = inPattern.firstEntity(); 
    Expression succ = outPattern.toExpression(types,ents,env,interp); 
    Vector env2 = outPattern.allVariables(); 

    Attribute inatt = (Attribute) env.get(0);
    String mainvar = inatt.getName(); 
    BasicExpression inexp = new BasicExpression(mainvar);
    ante = ante.dereference(inexp); 
    succ = succ.dereference(inexp);  

    Attribute outatt = (Attribute) env2.get(0);
    String resultvar = outatt.getName(); 
    BasicExpression outexp = new BasicExpression(resultvar);

    if (actionBlock != null) 
    { String opname = name + "ActionBlockop"; 
      Vector pars = new Vector(); 
      pars.addAll(env); 
      pars.remove(0); 
      pars = VectorUtil.union(pars,env2); 
      pars.remove(inatt); 
      Statement code = actionBlock.dereference(inexp); 
      BehaviouralFeature op = new BehaviouralFeature(opname,pars,false,null);  
      // op.setDerived(true); 
      op.addStereotype("explicit"); 
      op.setActivity(code); 
      op.setPost(new BasicExpression("true"));   
      owner.addOperation(op); 
      op.setEntity(owner); 
      String opcall = "self." + opname + "(";
      Vector parnames = new Vector();  
      for (int i = 1; i < env.size(); i++) 
      { String par = ((Attribute) env.get(i)).getName();
        if (parnames.contains(par) || mainvar.equals(par)) { } 
        else 
        { parnames.add(par); } 
      }
      for (int j = 0; j < env2.size(); j++) 
      { String par = ((Attribute) env2.get(j)).getName();
        if (parnames.contains(par) || mainvar.equals(par)) { } 
        else 
        { parnames.add(par); } 
      }  
      for (int k = 0; k < parnames.size(); k++) 
      { opcall = opcall + parnames.get(k); 
        if (k < parnames.size() - 1)
        { opcall = opcall + ","; } 
      }    
      opcall = opcall + ")"; 
      BasicExpression invokeop = 
        new BasicExpression(opcall,0); 
      Expression efo = invokeop.checkIfSetExpression(); 
      efo.setEntity(owner); // ??
      if (efo instanceof BasicExpression) 
      { BasicExpression befo = (BasicExpression) efo; 
        if (befo.objectRef != null) 
        { befo.objectRef.setEntity(owner); } 
      } 
      succ = Expression.simplify("&",succ,efo,null); 
    } 

    String opnme = name + "op"; 
    Vector pars0 = new Vector(); 
    pars0.addAll(env); 
    pars0.remove(0); 
    BehaviouralFeature op0 = new BehaviouralFeature(opnme,pars0,false,outatt.getType());  
    // op0.setDerived(true); 
    op0.addStereotype("explicit"); 
    op0.setPre(ante); 

    BinaryExpression setresult = new BinaryExpression("=",new BasicExpression("result"),outexp); 
    succ = Expression.simplify("&",succ,setresult,null);                              

    for (int i = 0; i < env2.size(); i++) 
    { Attribute att = (Attribute) env2.get(i);
      String nme = att.getName(); 
      String typ = att.getType() + "";  
      ModelElement met = ModelElement.lookupByName(att.getName(), env); 
      if (met == null)  // att is created by out pattern
      { succ = new BinaryExpression("#", new BinaryExpression(":",
                                        new BasicExpression(nme), new BasicExpression(typ)),
                                      succ);
      } // else, postfix each att.f by @pre in succ
      // else 
      // { succ = succ.addPreForms(att.getName()); } 
    }
    op0.setPost(succ);  // result = par0 
    owner.addOperation(op0); 
    op0.setEntity(owner); 
  } 

  public Constraint toConstraint(Vector types, Vector ents, java.util.Map interp)
  { Expression ante = inPattern.toExpression(); 
    if (ante == null) 
    { ante = new BasicExpression("true"); } 

    Vector env = inPattern.allVariables(); 
    Entity owner = inPattern.firstEntity(); 
    Expression succ = outPattern.toExpression(types,ents,env,interp); 
    Vector env2 = outPattern.allVariables(); 

    Attribute inatt = (Attribute) env.get(0);
    String mainvar = inatt.getName(); 
    BasicExpression inexp = new BasicExpression(mainvar);
    ante = ante.dereference(inexp); 
    succ = succ.dereference(inexp);  

    if (actionBlock != null) // If a rule is split, put this in the 2nd rule. 
    { String opname = name + "op"; 
      Vector pars = new Vector(); 
      pars.addAll(env); 
      pars.remove(0); 
      pars = VectorUtil.union(pars,env2); 
      pars.remove(inatt); 
      Statement code = actionBlock.dereference(inexp); 
      code.setEntity(owner); // self in code is an instance of owner
      BehaviouralFeature op = new BehaviouralFeature(opname,pars,false,null);  
      // op.setDerived(true); 
      op.addStereotype("explicit"); 
      op.setActivity(code); 
      owner.addOperation(op); 
      op.setPost(new BasicExpression("true"));   
      op.setEntity(owner); 
      String opcall = "self." + opname + "(";
      Vector parnames = new Vector();  
      for (int i = 1; i < env.size(); i++) 
      { String par = ((Attribute) env.get(i)).getName();
        if (parnames.contains(par) || mainvar.equals(par)) { } 
        else 
        { parnames.add(par); } 
      }
      for (int j = 0; j < env2.size(); j++) 
      { String par = ((Attribute) env2.get(j)).getName();
        if (parnames.contains(par) || mainvar.equals(par)) { } 
        else 
        { parnames.add(par); } 
      }  
      for (int k = 0; k < parnames.size(); k++) 
      { opcall = opcall + parnames.get(k); 
        if (k < parnames.size() - 1)
        { opcall = opcall + ","; } 
      }    
      opcall = opcall + ")"; 
      BasicExpression invokeop = 
        new BasicExpression(opcall,0); 
      Expression efo = invokeop.checkIfSetExpression(); 
      efo.setEntity(owner); // ??
      if (efo instanceof BasicExpression) 
      { BasicExpression befo = (BasicExpression) efo; 
        if (befo.objectRef != null) 
        { befo.objectRef.setEntity(owner); } 
      } 
      succ = Expression.simplify("&",succ,efo,null); 
    } 

    for (int i = 0; i < env2.size(); i++) 
    { Attribute att = (Attribute) env2.get(i);
      String nme = att.getName(); 
      String typ = att.getType() + "";  
      ModelElement met = ModelElement.lookupByName(att.getName(), env); 
      if (met == null)  // att is created by out pattern
      { succ = new BinaryExpression("#", new BinaryExpression(":",
                                        new BasicExpression(nme), new BasicExpression(typ)),
                                      succ);
      } // else, postfix each att.f by @pre in succ
      // else 
      // { succ = succ.addPreForms(att.getName()); } 
    }

    Constraint con = new Constraint(ante,succ); 
    con.setOwner(owner); 
    return con; 
  } 

  public boolean hasOutputType(Vector datas)
  { for (int i = 0; i < datas.size(); i++) 
    { String typ = (String) datas.get(i); 
      if (outPattern.hasType(typ)) 
      { return true; } 
    } 
    return false; 
  } 

  public MatchedRule slice()
  { MatchedRule newrule = new MatchedRule(isLazy,isUnique); 
    newrule.setName(name + "$1"); 
    newrule.setInPattern(inPattern); 
    OutPattern op = outPattern.slice(); 
    newrule.setOutPattern(op); 
    newrule.setActionBlock(actionBlock); 
    actionBlock = null; // put into newrule
    return newrule; 
  } 

} 

  