#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#define ALLOCATIONSIZE 10

#define TRUE 1
#define FALSE 0

#include "ocl.h"

struct A {
  int x;
  struct A** neighbours;
};

struct B {
  int y;
};

struct A** a_instances = NULL;
int a_size = 0;

struct B** b_instances = NULL;
int b_size = 0;

int getA_x(struct A* self) { return self->x; }

struct A** getA_neighbours(struct A* self) { return self->neighbours; }

int getB_y(struct B* self) { return self->y; }

void setA_x(struct A* self, int _value) { self->x = _value; }

void setA_neighbours(struct A* self, struct A** _value) { self->neighbours = _value; }

void setB_y(struct B* self, int _value) { self->y = _value; }

int* getAllA_x(struct A* _col[])
{ int n = length((void**) _col);
  int* result = (int*) calloc(n, sizeof(int));
  int i = 0;
  for ( ; i < n; i++)
  { result[i] = getA_x(_col[i]); }
  return result;
}

int* getAllB_y(struct B* _col[])
{ int n = length((void**) _col);
  int* result = (int*) calloc(n, sizeof(int));
  int i = 0;
  for ( ; i < n; i++)
  { result[i] = getB_y(_col[i]); }
  return result;
}

struct A** appendA(struct A* col[], struct A* ex)
   { struct A** result;
     int len = length((void**) col);
     if (len % ALLOCATIONSIZE == 0)
     { result = (struct A**) calloc(len + ALLOCATIONSIZE + 1, sizeof(struct A*));
       int i = 0;
       for ( ; i < len; i++) { result[i] = col[i]; }
     }
    else { result = col; }
    result[len] = ex;
    result[len+1] = NULL;
    return result;
  }

struct B** appendB(struct B* col[], struct B* ex)
   { struct B** result;
     int len = length((void**) col);
     if (len % ALLOCATIONSIZE == 0)
     { result = (struct B**) calloc(len + ALLOCATIONSIZE + 1, sizeof(struct B*));
       int i = 0;
       for ( ; i < len; i++) { result[i] = col[i]; }
     }
    else { result = col; }
    result[len] = ex;
    result[len+1] = NULL;
    return result;
  }

struct A* createA()
{ struct A* result = (struct A*) malloc(sizeof(struct A));
  result->x = 0;
  result->neighbours = NULL;
  a_instances = appendA(a_instances, result);
  a_size++;
  return result;
}

struct B* createB()
{ struct B* result = (struct B*) malloc(sizeof(struct B));
  result->y = 0;
  b_instances = appendB(b_instances, result);
  b_size++;
  return result;
}

struct A** newAList()
{ return (struct A**) calloc(ALLOCATIONSIZE, sizeof(struct A*)); }

struct B** newBList()
{ return (struct B**) calloc(ALLOCATIONSIZE, sizeof(struct B*)); }

struct A** insertA(struct A* col[], struct A* self)
  { if (isIn((void*) self, (void**) col))
    { return col; }
    return appendA(col,self);
  }

struct B** insertB(struct B* col[], struct B* self)
  { if (isIn((void*) self, (void**) col))
    { return col; }
    return appendB(col,self);
  }

struct A** selectA(struct A* col[], int (*test)(struct A* self))
{ struct A** result;
  int len = length((void**) col);
  result = (struct A**) calloc(len + 1, sizeof(struct A*));
  int j = 0;
  int i = 0;
  for ( ; i < len; i++)
  { struct A* self = col[i];
    if (self == NULL)
	{ result[j] = NULL;
	  return result;
	}
	if ((*test)(self))
    { result[j] = self;
	  j++;
	}
  }
  result[j] = NULL;
  return result;
}

struct B** selectB(struct B* col[], int (*test)(struct B* self))
{ struct B** result;
  int len = length((void**) col);
  result = (struct B**) calloc(len + 1, sizeof(struct B*));
  int j = 0;
  int i = 0;
  for ( ; i < len; i++)
  { struct B* self = col[i];
    if (self == NULL)
	{ result[j] = NULL;
	  return result;
	}
	if ((*test)(self))
    { result[j] = self;
	  j++;
	}
  }
  result[j] = NULL;
  return result;
}

struct A** rejectA(struct A* col[], int (*test)(struct A* self))
{ struct A** result;
  int len = length((void**) col);
  result = (struct A**) calloc(len + 1, sizeof(struct A*));
  int j = 0;
  int i = 0;
  for ( ; i < len; i++)
  { struct A* self = col[i];
    if (self == NULL)
	{ result[j] = NULL;
	  return result;
	}
	if ((*test)(self)) {}
    else { result[j] = self;
	  j++;
	}
  }
  result[j] = NULL;
  return result;
}

struct B** rejectB(struct B* col[], int (*test)(struct B* self))
{ struct B** result;
  int len = length((void**) col);
  result = (struct B**) calloc(len + 1, sizeof(struct B*));
  int j = 0;
  int i = 0;
  for ( ; i < len; i++)
  { struct B* self = col[i];
    if (self == NULL)
	{ result[j] = NULL;
	  return result;
	}
	if ((*test)(self)) {}
    else { result[j] = self;
	  j++;
	}
  }
  result[j] = NULL;
  return result;
}

   void** collectA(struct A* col[], void* (*fe)(struct A*))
   { int n = length((void**) col);
    void** result = (void**) calloc(n+1, sizeof(void*));
    int i = 0;
    for ( ; i < n; i++)
    { struct A* self = col[i];
      result[i] = (*fe)(self);
    }
    result[n] = NULL;
    return result;
  }

   void** collectB(struct B* col[], void* (*fe)(struct B*))
   { int n = length((void**) col);
    void** result = (void**) calloc(n+1, sizeof(void*));
    int i = 0;
    for ( ; i < n; i++)
    { struct B* self = col[i];
      result[i] = (*fe)(self);
    }
    result[n] = NULL;
    return result;
  }

 unsigned char existsA(struct A* col[], int (*test)(struct A* ex))
 { int n = length((void**) col);
   unsigned char result = FALSE;
   int i = 0;
   for ( ; i < n; i++)
   { struct A* ex = col[i];
     if (ex == NULL) { return result; }
     if ((*test)(ex))
     { return TRUE; }
   }
   return result;
 }

 unsigned char existsB(struct B* col[], int (*test)(struct B* ex))
 { int n = length((void**) col);
   unsigned char result = FALSE;
   int i = 0;
   for ( ; i < n; i++)
   { struct B* ex = col[i];
     if (ex == NULL) { return result; }
     if ((*test)(ex))
     { return TRUE; }
   }
   return result;
 }

 unsigned char forAllA(struct A* col[], int (*test)(struct A* ex))
 { int n = length((void**) col);
   unsigned char result = TRUE;
   int i = 0;
   for ( ; i < n; i++)
   { struct A* ex = col[i];
     if (ex == NULL) { return result; }
     if ((*test)(ex)) {}
     else { return FALSE; }
   }
   return result;
 }

 unsigned char forAllB(struct B* col[], int (*test)(struct B* ex))
 { int n = length((void**) col);
   unsigned char result = TRUE;
   int i = 0;
   for ( ; i < n; i++)
   { struct B* ex = col[i];
     if (ex == NULL) { return result; }
     if ((*test)(ex)) {}
     else { return FALSE; }
   }
   return result;
 }

  struct A** subrangeA(struct A** col, int i, int j)
  { int len = length((void**) col);
    if (i > j || j > len) { return NULL; }
    struct A** result = (struct A**) calloc(j - i + 2, sizeof(struct A*));
     int k = i-1;
     int l = 0;
     for ( ; k < j; k++, l++)
     { result[l] = col[k]; }
    result[l] = NULL;
    return result;
  }

  struct B** subrangeB(struct B** col, int i, int j)
  { int len = length((void**) col);
    if (i > j || j > len) { return NULL; }
    struct B** result = (struct B**) calloc(j - i + 2, sizeof(struct B*));
     int k = i-1;
     int l = 0;
     for ( ; k < j; k++, l++)
     { result[l] = col[k]; }
    result[l] = NULL;
    return result;
  }

  struct A** reverseA(struct A** col)
  { int n = length((void**) col);
    struct A** result = (struct A**) calloc(n+1, sizeof(struct A*));
    int i = 0;
    int x = n-1;
    for ( ; i < n; i++, x--)
    { result[i] = col[x]; }
    result[n] = NULL;
    return result;
  }

  struct B** reverseB(struct B** col)
  { int n = length((void**) col);
    struct B** result = (struct B**) calloc(n+1, sizeof(struct B*));
    int i = 0;
    int x = n-1;
    for ( ; i < n; i++, x--)
    { result[i] = col[x]; }
    result[n] = NULL;
    return result;
  }

struct A** removeA(struct A* col[], struct A* ex)
{ int len = length((void**) col);
  struct A** result = (struct A**) calloc(len+1, sizeof(struct A*));
  int j = 0;
  int i = 0;
  for ( ; i < len; i++)
  { struct A* eobj = col[i];
    if (eobj == NULL)
    { result[j] = NULL;
      return result; 
    }
    if (eobj == ex) { }
    else
    { result[j] = eobj; j++; }
  }
  result[j] = NULL;
  return result;
}

struct B** removeB(struct B* col[], struct B* ex)
{ int len = length((void**) col);
  struct B** result = (struct B**) calloc(len+1, sizeof(struct B*));
  int j = 0;
  int i = 0;
  for ( ; i < len; i++)
  { struct B* eobj = col[i];
    if (eobj == NULL)
    { result[j] = NULL;
      return result; 
    }
    if (eobj == ex) { }
    else
    { result[j] = eobj; j++; }
  }
  result[j] = NULL;
  return result;
}

struct A** unionA(struct A* col1[], struct A* col2[])
  { int n = length((void**) col1);
    int m = length((void**) col2);
    struct A** result = (struct A**) calloc(n + m + 1, sizeof(struct A*));
    int i = 0;
    int j = 0;
    for ( ; i < n; i++)
    { struct A* ex = col1[i];
      if (isIn((void*) ex, (void**) result)) { }
      else   
      { result[j] = ex; j++; }
    }    
    i = 0;
    for ( ; i < m; i++)
    { struct A* ex = col2[i];
      if (isIn((void*) ex, (void**) result)) { }
      else   
      { result[j] = ex; j++; }
    }
    result[j] = NULL;
    return result;
  }

struct B** unionB(struct B* col1[], struct B* col2[])
  { int n = length((void**) col1);
    int m = length((void**) col2);
    struct B** result = (struct B**) calloc(n + m + 1, sizeof(struct B*));
    int i = 0;
    int j = 0;
    for ( ; i < n; i++)
    { struct B* ex = col1[i];
      if (isIn((void*) ex, (void**) result)) { }
      else   
      { result[j] = ex; j++; }
    }    
    i = 0;
    for ( ; i < m; i++)
    { struct B* ex = col2[i];
      if (isIn((void*) ex, (void**) result)) { }
      else   
      { result[j] = ex; j++; }
    }
    result[j] = NULL;
    return result;
  }

struct A** removeAllA(struct A* col1[], struct A* col2[])
{ int n = length((void**) col1);
  struct A** result = (struct A**) calloc(n+1, sizeof(struct A*));
  int i = 0; int j = 0;
  for ( ; i < n; i++)
  { struct A* ex = col1[i];
    if (isIn((void*) ex, (void**) col2)) {}
    else 
    { result[j] = ex; j++; }
  }
  result[j] = NULL;
  return result;
}

struct B** removeAllB(struct B* col1[], struct B* col2[])
{ int n = length((void**) col1);
  struct B** result = (struct B**) calloc(n+1, sizeof(struct B*));
  int i = 0; int j = 0;
  for ( ; i < n; i++)
  { struct B* ex = col1[i];
    if (isIn((void*) ex, (void**) col2)) {}
    else 
    { result[j] = ex; j++; }
  }
  result[j] = NULL;
  return result;
}

struct A** frontA(struct A* col[])
{ int n = length((void**) col);
  return subrangeA(col, 1, n-1); }

struct B** frontB(struct B* col[])
{ int n = length((void**) col);
  return subrangeB(col, 1, n-1); }

struct A** tailA(struct A* col[])
{ int n = length((void**) col);
  return subrangeA(col, 2, n); }

struct B** tailB(struct B* col[])
{ int n = length((void**) col);
  return subrangeB(col, 2, n); }

struct A** concatenateA(struct A* _col1[], struct A* _col2[])
{ int n = length((void**) _col1);
  int m = length((void**) _col2);
  struct A** result = (struct A**) calloc(n + m + 1, sizeof(struct A*));
  int i = 0;
  int j = 0;
  for ( ; i < n; i++)
  { result[j] = _col1[i];
    j++;
  }
  i = 0;
  for ( ; i < m; i++)
  { result[j] = _col2[i];
    j++;
  }
  result[j] = NULL;
  return result;
}

struct B** concatenateB(struct B* _col1[], struct B* _col2[])
{ int n = length((void**) _col1);
  int m = length((void**) _col2);
  struct B** result = (struct B**) calloc(n + m + 1, sizeof(struct B*));
  int i = 0;
  int j = 0;
  for ( ; i < n; i++)
  { result[j] = _col1[i];
    j++;
  }
  i = 0;
  for ( ; i < m; i++)
  { result[j] = _col2[i];
    j++;
  }
  result[j] = NULL;
  return result;
}

struct A** intersectionA(struct A* _col1[], struct A* _col2[])
{ int n = length((void**) _col1);
  int m = length((void**) _col2);
  struct A** result = (struct A**) calloc(n + 1, sizeof(struct A*));
  int i = 0;
  int j = 0;
  for ( ; i < n; i++)
  { struct A* _ex = _col1[i];
    if (isIn((void*) _ex, (void**) _col2))
    { result[j] = _ex; j++; }
  }
  result[j] = NULL;
  return result;
}

struct B** intersectionB(struct B* _col1[], struct B* _col2[])
{ int n = length((void**) _col1);
  int m = length((void**) _col2);
  struct B** result = (struct B**) calloc(n + 1, sizeof(struct B*));
  int i = 0;
  int j = 0;
  for ( ; i < n; i++)
  { struct B* _ex = _col1[i];
    if (isIn((void*) _ex, (void**) _col2))
    { result[j] = _ex; j++; }
  }
  result[j] = NULL;
  return result;
}

struct A** insertAtA(struct A* col1[], int ind, struct A* col2[])
{ if (ind <= 0) { return col1; }
  int n = length((void**) col1);
  int m = length((void**) col2);
  if (m == 0) { return col1; }
  struct A** result = (struct A**) calloc(n + m + 1, sizeof(struct A*));
  int i = 0; int j = 0;
  for ( ; i < ind - 1 && i < n; i++)
  { result[i] = col1[i]; }
  if (i == ind - 1)
  { for ( ; j < m; j++, i++)
    { result[i] = col2[j]; }
    for ( ; i < n + m; i++)
    { result[i] = col1[i - m]; }
  }
  else 
  { for ( ; j < m; j++, i++)
    { result[i] = col2[j]; }
  }
  result[n+m] = NULL;
  return result;
}

struct B** insertAtB(struct B* col1[], int ind, struct B* col2[])
{ if (ind <= 0) { return col1; }
  int n = length((void**) col1);
  int m = length((void**) col2);
  if (m == 0) { return col1; }
  struct B** result = (struct B**) calloc(n + m + 1, sizeof(struct B*));
  int i = 0; int j = 0;
  for ( ; i < ind - 1 && i < n; i++)
  { result[i] = col1[i]; }
  if (i == ind - 1)
  { for ( ; j < m; j++, i++)
    { result[i] = col2[j]; }
    for ( ; i < n + m; i++)
    { result[i] = col1[i - m]; }
  }
  else 
  { for ( ; j < m; j++, i++)
    { result[i] = col2[j]; }
  }
  result[n+m] = NULL;
  return result;
}

unsigned char exists1A(struct A* col[], int (*test)(struct A* ex))
{ int n = length((void**) col);
  unsigned char result = FALSE; 
  unsigned char found = FALSE;
  int i = 0;  
  for ( ; i < n; i++)
  { struct A* ex = col[i];
    if (ex == NULL) { return result; }
    if ((*test)(ex))
    { if (found) { return FALSE; }
      else { found = TRUE; }
    }
  }
  if (found) { return TRUE; }
  return result;
}

unsigned char exists1B(struct B* col[], int (*test)(struct B* ex))
{ int n = length((void**) col);
  unsigned char result = FALSE; 
  unsigned char found = FALSE;
  int i = 0;  
  for ( ; i < n; i++)
  { struct B* ex = col[i];
    if (ex == NULL) { return result; }
    if ((*test)(ex))
    { if (found) { return FALSE; }
      else { found = TRUE; }
    }
  }
  if (found) { return TRUE; }
  return result;
}

unsigned char isUniqueA(struct A* col[], void* (*fe)(struct A* ex))
{ unsigned char result = TRUE; 
  void** values = collectA(col, fe);
  int n = length((void**) values);
  int i = 0;  
  for ( ; i < n; i++)
  { void* ex = values[i];
    if (i < n - 1 && isIn(ex, values + (i + 1)))
    { return FALSE; }
  }
  return result;
}

unsigned char isUniqueB(struct B* col[], void* (*fe)(struct B* ex))
{ unsigned char result = TRUE; 
  void** values = collectB(col, fe);
  int n = length((void**) values);
  int i = 0;  
  for ( ; i < n; i++)
  { void* ex = values[i];
    if (i < n - 1 && isIn(ex, values + (i + 1)))
    { return FALSE; }
  }
  return result;
}

struct A** asSetA(struct A* col[])
{ int n = length((void**) col);
  if (n == 0) { return col; }
  struct A** result = (struct A**) calloc(n + 1, sizeof(struct A*));
  int i = 0; int j = 0;
  result[j] = NULL;
  for ( ; i < n; i++)
  { struct A* ex = col[i];
    if (isIn((void*) ex, (void**) result)) {}
    else  
    { result[j] = col[i]; j++;
      result[j] = NULL; }
  }    
  result[j] = NULL;
  return result;
}

struct B** asSetB(struct B* col[])
{ int n = length((void**) col);
  if (n == 0) { return col; }
  struct B** result = (struct B**) calloc(n + 1, sizeof(struct B*));
  int i = 0; int j = 0;
  result[j] = NULL;
  for ( ; i < n; i++)
  { struct B* ex = col[i];
    if (isIn((void*) ex, (void**) result)) {}
    else  
    { result[j] = col[i]; j++;
      result[j] = NULL; }
  }    
  result[j] = NULL;
  return result;
}

